import functools
from abc import ABC, abstractmethod

import requests

from .exceptions import (
    SourceSite5xxError,
    SourceSiteTimeoutError,
    SourceSiteUnexpectedError,
)


class Site(ABC):
    def __init__(self):
        self.username = None
        self.password = None
        self.is_logged_in = False
        self._session = requests.Session()
        self._session.headers.update(
            {
                "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/78.0.3904.108 Safari/537.36"
            }
        )

    @abstractmethod
    def get_title(self, title_id):
        pass

    @abstractmethod
    def get_chapter(self, title_id, chapter_id):
        pass

    @abstractmethod
    def search_title(self, query):
        pass

    @abstractmethod
    def title_cover(self, title_id, cover_ext):
        pass

    @abstractmethod
    def title_thumbnail(self, title_id):
        pass

    @abstractmethod
    def title_source_url(self, title_id):
        pass

    # optional abstract method
    def login(self, username, password):
        raise NotImplementedError()

    def _http_request(self, method, url, *args, **kwargs):
        request_func = getattr(self._session, method)

        if "timeout" not in kwargs:
            kwargs["timeout"] = 5

        try:
            resp = request_func(url, *args, **kwargs)
        except requests.exceptions.Timeout:
            raise SourceSiteTimeoutError(url)

        if 500 <= resp.status_code <= 599:
            raise SourceSite5xxError(url, resp.status_code, resp.text)
        elif resp.status_code != 200:
            raise SourceSiteUnexpectedError(url, resp.status_code, resp.text)

        return resp

    def http_get(self, *args, **kwargs):
        return self._http_request("get", *args, **kwargs)

    def http_post(self, *args, **kwargs):
        return self._http_request("post", *args, **kwargs)


def requires_login(func):
    """
    Decorator designed for use on a Site's instance methods.
    It ensures cookies are ready before running the method.
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        # TODO: replace is_logged_in flag check with actual "if rejected then try
        # logging in" logic, just in case login cookies expire.
        if not self.is_logged_in:
            assert self.username
            assert self.password
            self.login(self.username, self.password)
        return func(self, *args, **kwargs)

    return wrapper



class SourceSiteResponseError(Exception):
    def __init__(self, url, status_code=None, response_text=None):
        self.url = url
        self.status_code = status_code
        self.response_text = response_text


class SourceSiteTimeoutError(SourceSiteResponseError):
    pass


class SourceSite5xxError(SourceSiteResponseError):
    pass


class SourceSiteUnexpectedError(SourceSiteResponseError):
    pass



import html
import re
import time

from mangoapi.base_site import Site, requires_login


class Mangadex(Site):
    def get_title(self, title_id):
        url = f"https://mangadex.org/api/?id={title_id}&type=manga"
        md_resp = self.http_get(url)
        md_json = md_resp.json()
        assert md_json["status"] == "OK"

        cover = md_json["manga"]["cover_url"].split("/")[-1]
        cover_ext = cover[cover.find(".") + 1 : cover.rfind("?")]

        current_timestamp = time.time()

        title = {
            "id": title_id,
            "name": md_json["manga"]["title"],
            "site": "mangadex",
            "cover_ext": cover_ext,
            "alt_names": md_json["manga"]["alt_names"],
            "descriptions": html.unescape(md_json["manga"]["description"]).split(
                "\r\n\r\n"
            ),
            "chapters": [
                {
                    "id": str(chap_id),
                    "name": chap["title"],
                    "volume": int(chap["volume"]) if chap["volume"] else None,
                    "groups": _extract_groups(chap),
                    **_parse_chapter_number(chap["chapter"]),
                }
                for chap_id, chap in md_json.get("chapter", {}).items()
                if chap["lang_code"] == "gb"
                and chap["group_name"] != "MangaPlus"
                and chap["timestamp"] <= current_timestamp
                # ^ Chapter may be listed but with access delayed for a certain amount
                # of time set by uploader, in which case we just filter it out. God I
                # hate this generation of Patreon "scanlators".
            ],
        }
        return title

    def get_chapter(self, title_id, chapter_id):
        md_resp = self.http_get(
            f"https://mangadex.org/api/?id={chapter_id}&type=chapter&saver=0"
        )
        md_json = md_resp.json()
        assert md_json["status"] == "OK"

        # 2 cases:
        # - If 'server_fallback' is absent, it means 'server' points to MD's own server
        #   e.g. s5.mangadex.org...
        # - Otherwise, 'server' points to a likely ephemeral MD@H node, while
        # 'server_fallback' now points to MD's own server.
        #
        # MD's own links apparently go dead sometimes, but MD@H links seem to expire
        # quickly all the time, so it's probably a good idea to store both anyway.

        server_fallback = md_json.get("server_fallback")
        if server_fallback:
            md_server = server_fallback
            mdah_server = md_json["server"]
        else:
            md_server = md_json["server"]
            mdah_server = None

        chapter = {
            "id": chapter_id,
            "title_id": str(md_json["manga_id"]),
            "site": "mangadex",
            "name": md_json["title"],
            "pages": [
                f"{md_server}{md_json['hash']}/{page}" for page in md_json["page_array"]
            ],
            "pages_alt": [
                f"{mdah_server}{md_json['hash']}/{page}"
                for page in md_json["page_array"]
            ]
            if mdah_server
            else [],
            "groups": _extract_groups(md_json),
            "is_webtoon": md_json["long_strip"] == 1,
            **_parse_chapter_number(md_json["chapter"]),
        }
        return chapter

    @requires_login
    def search_title(self, query):
        md_resp = self.http_get(f"https://mangadex.org/quick_search/{query}")

        matches = TITLES_PATTERN.findall(md_resp.text)
        titles = [
            {
                "id": id,
                "name": name.strip(),
                "site": "mangadex",
                "thumbnail": f"https://mangadex.org/images/manga/{id}.large.jpg",
            }
            for id, name in matches
        ]
        return titles

    def login(self, username, password):
        form_data = {
            "login_username": username,
            "login_password": password,
            "two_factor": "",
            "remember_me": "1",
        }
        self.http_post(
            "https://mangadex.org/ajax/actions.ajax.php?function=login",
            data=form_data,
            headers={"X-Requested-With": "XMLHttpRequest"},
        )
        self.is_logged_in = True

    def title_cover(self, title_id, cover_ext):
        return f"https://mangadex.org/images/manga/{title_id}.{cover_ext}"

    def title_thumbnail(self, title_id):
        return f"https://mangadex.org/images/manga/{title_id}.large.jpg"

    def title_source_url(self, title_id):
        return f"https://mangadex.org/manga/{title_id}"


# Titles regex slightly adapted from https://github.com/md-y/mangadex-full-api
# Thanks!
TITLES_PATTERN = re.compile(
    r"""<a[^>]*href=["']\/title\/(\d+)\/\S+["'][^>]*manga_title[^>]*>([^<]*)<"""
)


def _parse_chapter_number(string):
    if string == "":
        # most likely a oneshot
        return {"number": ""}
    nums = string.split(".")
    count = len(nums)
    assert count == 1 or count == 2
    result = {"number": string}
    result["num_major"] = int(nums[0])
    if count == 2:
        result["num_minor"] = int(nums[1])
    return result


def _extract_groups(chap):
    return [
        html.unescape(group.strip())
        for group in [chap["group_name"], chap["group_name_2"], chap["group_name_3"]]
        if group
    ]



from secrets import token_urlsafe

from goodconf import GoodConf, Value


class Config(GoodConf):
    MANGADEX_USERNAME = Value()
    MANGADEX_PASSWORD = Value()
    FLASK_SECRET_KEY = Value(initial=lambda: token_urlsafe(50))

    PROXY_PREFIX = Value(default="")
    # ^ use this to e.g. point to a CDN's domain
    PROXY_CACHE_DIR = Value(default="proxy_cache")
    PROXY_CACHE_MAX_SIZE = Value(default=1024 * 1024 * 1024 * 5)  # 5GiB in bytes
    PROXY_CACHE_MAX_AGE = Value(default=3600 * 24 * 2)  # 2 days in seconds


config = Config(default_files=["pytaku.conf.json"])



from .mangadex import Mangadex
from .mangasee import Mangasee

"""
The mangoapi package is designed to be self-contained as if it was an external library.
Each Site object represents a user session on source site.
Instantiating and managing Site objects is the responsibility of the caller.
"""

SITES = {
    "mangadex": Mangadex,
    "mangasee": Mangasee,
}


def get_site_class(name):
    return SITES.get(name)



from functools import wraps

from flask import jsonify, render_template, request

from mangoapi.exceptions import (
    SourceSite5xxError,
    SourceSiteResponseError,
    SourceSiteTimeoutError,
    SourceSiteUnexpectedError,
)

from .persistence import verify_token


def process_token(required=True):
    def decorator(f):
        @wraps(f)
        def decorated_function(*args, **kwargs):
            token = request.cookies.get("token")
            if not token:
                if required:
                    return (
                        jsonify({"message": "Missing 'token' cookie."}),
                        401,
                    )
                else:
                    request.token = None
                    request.user_id = None
                    return f(*args, **kwargs)

            user_id = verify_token(token)
            if user_id is None:
                return jsonify({"message": "Invalid token."}), 401

            request.token = token
            request.user_id = user_id

            return f(*args, **kwargs)

        return decorated_function

    return decorator


SOURCE_SITE_ERRORS = {
    SourceSiteTimeoutError: {
        "code": "source_site_timeout",
        "message": "Source site took too long to respond. Try again later.",
    },
    SourceSite5xxError: {
        "code": "source_site_5xx",
        "message": "Source site crapped the bed. Try again later.",
    },
    SourceSiteUnexpectedError: {
        "code": "source_site_unexpected",
        "message": "Unexpected error when requesting source site.",
    },
}


def handle_source_site_errors(format="json"):
    assert format in ("json", "html"), f"{format} ain't no format I ever heard of!"

    def outer_func(f):
        @wraps(f)
        def decorated_function(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except SourceSiteResponseError as err:
                resp = SOURCE_SITE_ERRORS[err.__class__]
                resp["detail"] = err.__dict__
                if format == "json":
                    return jsonify(resp), 500
                elif format == "html":
                    return render_template("error.html", **resp), 500

        return decorated_function

    return outer_func



import json
import secrets
from typing import List, Tuple

import apsw
import argon2

from .database.common import run_sql, run_sql_many


def save_title(title):
    run_sql(
        """
    INSERT INTO title (
        id,
        name,
        site,
        cover_ext,
        chapters,
        alt_names,
        descriptions
    ) VALUES (
        :id,
        :name,
        :site,
        :cover_ext,
        :chapters,
        :alt_names,
        :descriptions
    ) ON CONFLICT (id, site) DO UPDATE SET
        name=excluded.name,
        cover_ext=excluded.cover_ext,
        chapters=excluded.chapters,
        alt_names=excluded.alt_names,
        descriptions=excluded.descriptions,
        updated_at=datetime('now')
    ;
    """,
        {
            "id": title["id"],
            "name": title["name"],
            "site": title["site"],
            "cover_ext": title["cover_ext"],
            "chapters": json.dumps(title["chapters"]),
            "alt_names": json.dumps(title["alt_names"]),
            "descriptions": json.dumps(title["descriptions"]),
        },
    )


def load_title(site, title_id, user_id=None):
    result = run_sql(
        """
        SELECT id, name, site, cover_ext, chapters, alt_names, descriptions
        FROM title
        WHERE id = ?
          AND site = ?;
        """,
        (title_id, site),
    )
    if not result:
        return None
    elif len(result) > 1:
        raise Exception(f"Found multiple results for title_id {title_id} on {site}!")
    else:
        title = result[0]

        for field in ["chapters", "alt_names", "descriptions"]:
            title[field] = json.loads(title[field])

        if user_id is not None:
            title["is_following"] = bool(
                run_sql(
                    "SELECT 1 FROM follow WHERE user_id=? AND site=? AND title_id=?;",
                    (user_id, site, title["id"]),
                )
            )

            chapters_i_read = run_sql(
                """
                SELECT r.chapter_id
                FROM read r
                WHERE r.user_id = ?
                  AND r.title_id = ?
                  AND r.site = ?
                ORDER BY r.updated_at;
                """,
                (user_id, title["id"], title["site"]),
            )

            for ch in title["chapters"]:
                if ch["id"] in chapters_i_read:
                    ch["is_read"] = True
        return title


def save_chapter(chapter):
    run_sql(
        """
    INSERT INTO chapter (
        id,
        title_id,
        site,
        num_major,
        num_minor,
        name,
        pages,
        pages_alt,
        groups,
        is_webtoon
    ) VALUES (
        :id,
        :title_id,
        :site,
        :num_major,
        :num_minor,
        :name,
        :pages,
        :pages_alt,
        :groups,
        :is_webtoon
    ) ON CONFLICT (id, title_id, site) DO UPDATE SET
        num_major=excluded.num_major,
        num_minor=excluded.num_minor,
        name=excluded.name,
        pages=excluded.pages,
        pages_alt=excluded.pages_alt,
        groups=excluded.groups,
        is_webtoon=excluded.is_webtoon,
        updated_at=datetime('now')
    ;
    """,
        {
            "id": chapter["id"],
            "title_id": chapter["title_id"],
            "site": chapter["site"],
            "num_major": chapter.get("num_major"),
            "num_minor": chapter.get("num_minor"),
            "name": chapter["name"],
            "pages": json.dumps(chapter["pages"]),
            "pages_alt": json.dumps(chapter["pages_alt"]),
            "groups": json.dumps(chapter["groups"]),
            "is_webtoon": chapter["is_webtoon"],
        },
    )


def load_chapter(site, title_id, chapter_id, ignore_old=True):
    updated_at = "datetime('now', '-1 days')" if ignore_old else "'1980-01-01'"
    result = run_sql(
        f"""
        SELECT id, title_id, site, num_major, num_minor, name, pages, pages_alt, groups, is_webtoon
        FROM chapter
        WHERE site=? AND title_id=? AND id=? AND updated_at > {updated_at};
        """,
        (site, title_id, chapter_id),
    )
    if not result:
        return None
    elif len(result) > 1:
        raise Exception(f"Found multiple results for chapter_id {chapter_id}!")
    else:
        chapter = result[0]
        chapter["pages"] = json.loads(chapter["pages"])
        chapter["pages_alt"] = json.loads(chapter["pages_alt"])
        chapter["groups"] = json.loads(chapter["groups"])
        return chapter


def get_prev_next_chapters(title, chapter):
    """
    Maybe consider writing SQL query instead?
    """
    chapters = title["chapters"]
    chapter_id = chapter["id"]

    prev_chapter = None
    next_chapter = None
    for i, chap in enumerate(chapters):
        if chap["id"] == chapter_id:
            if i - 1 >= 0:
                next_chapter = chapters[i - 1]
            if i + 1 < len(chapters):
                prev_chapter = chapters[i + 1]

    return prev_chapter, next_chapter


def register_user(username, password):
    hasher = argon2.PasswordHasher()
    hashed_password = hasher.hash(password)
    try:
        run_sql(
            "INSERT INTO user (username, password) VALUES (?, ?);",
            (username, hashed_password),
        )
        return None
    except apsw.ConstraintError as e:
        if "UNIQUE" in str(e):
            return "Username already exists."
        raise


def verify_username_password(username, password):
    data = run_sql("SELECT id, password FROM user WHERE username = ?;", (username,))
    if len(data) != 1:
        print(f"User {username} doesn't exist.")
        return None

    user_id = data[0]["id"]
    hashed_password = data[0]["password"]
    hasher = argon2.PasswordHasher()
    try:
        hasher.verify(hashed_password, password)
        return user_id
    except argon2.exceptions.VerifyMismatchError:
        print(f"User {username} exists but password doesn't match.")
        return None


def follow(user_id, site, title_id):
    run_sql(
        """
        INSERT INTO follow (user_id, site, title_id) VALUES (?, ?, ?)
        ON CONFLICT DO NOTHING;
        """,
        (user_id, site, title_id),
    )


def unfollow(user_id, site, title_id):
    run_sql(
        "DELETE FROM follow WHERE user_id=? AND site=? AND title_id=?;",
        (user_id, site, title_id),
    )


def get_followed_titles(user_id):
    titles = run_sql(
        """
        SELECT t.id, t.site, t.name, t.cover_ext, t.chapters
        FROM title t
          INNER JOIN follow f ON f.title_id = t.id AND f.site = t.site
          INNER JOIN user u ON u.id = f.user_id
        WHERE user_id=?;
        """,
        (user_id,),
    )

    for t in titles:
        chapters = json.loads(t["chapters"])

        # n+1 queries cuz I don't give a f- actually I do, but sqlite's cool with it:
        # https://www.sqlite.org/np1queryprob.html
        chapters_i_finished = run_sql(
            """
            SELECT chapter_id
            FROM read
            WHERE user_id = ?
              AND title_id = ?
              AND site = ?;
            """,
            (user_id, t["id"], t["site"]),
        )
        # Only show chapters that user hasn't read
        chapters = [ch for ch in chapters if ch["id"] not in chapters_i_finished]

        t["chapters"] = chapters

    return sorted(titles, key=lambda t: len(t["chapters"]), reverse=True)


def read(user_id, site, title_id, chapter_id):
    run_sql(
        """
        INSERT INTO read (user_id, site, title_id, chapter_id) VALUES (?,?,?,?)
        ON CONFLICT (user_id, site, title_id, chapter_id)
        DO UPDATE SET updated_at=datetime('now');
        """,
        (user_id, site, title_id, chapter_id),
    )


def unread(user_id, site, title_id, chapter_id):
    run_sql(
        "DELETE FROM read WHERE user_id=? AND site=? AND title_id=? AND chapter_id=?;",
        (user_id, site, title_id, chapter_id),
    )


def find_outdated_titles(since="-6 hours"):
    return run_sql(
        "SELECT id, site FROM title WHERE updated_at <= datetime('now', ?);", (since,)
    )


class KeyvalStore:
    @staticmethod
    def get(key: str, default=None, since=None) -> str:
        if since is None:
            result = run_sql("SELECT value FROM keyval_store WHERE key=?;", (key,))
        else:
            result = run_sql(
                """
                SELECT value FROM keyval_store WHERE key=?
                AND updated_at >= datetime('now', ?);
                """,
                (key, since),
            )
        return result[0] if result else default

    @staticmethod
    def set(key: str, value: str):
        # let's not allow crap in by accident
        assert isinstance(key, str)
        assert isinstance(value, str)
        run_sql(
            """
            INSERT INTO keyval_store (key, value) VALUES (?,?)
            ON CONFLICT (key) DO UPDATE SET value=excluded.value, updated_at=datetime('now');
            """,
            (key, value),
        )


def import_follows(user_id: int, site_title_pairs: List[Tuple[str, str]]):
    run_sql_many(
        """
        INSERT INTO follow (user_id, site, title_id) VALUES (?, ?, ?)
        ON CONFLICT DO NOTHING;
        """,
        ((user_id, site, title_id) for site, title_id in site_title_pairs),
    )


def create_token(user_id, remember=False):
    lifespan = "+365 days" if remember else "+1 day"
    token = secrets.token_urlsafe(64)
    run_sql(
        """
        INSERT INTO token (user_id, token, lifespan) VALUES (?,?,?);
        """,
        (user_id, token, lifespan),
    )
    return token


def verify_token(token):
    """
    Checks if there's a matching token that hasn't exceeded its lifespan.
    If there's a match, refreshes its last_accessed_at value, effectively expanding
    its life.
    """
    result = run_sql(
        """
        SELECT user_id FROM token
        WHERE token=?
          AND datetime(last_accessed_at, lifespan) > datetime('now');
        """,
        (token,),
    )
    user_id = result[0] if len(result) == 1 else None
    if user_id:
        run_sql(
            "UPDATE token SET last_accessed_at = datetime('now') WHERE token=?;",
            (token,),
        )
    return user_id


def get_username(user_id):
    result = run_sql("SELECT username FROM user WHERE id=?;", (user_id,))
    assert len(result) == 1
    return result[0]


def delete_token(token):
    num_deleted = run_sql(
        "DELETE FROM token WHERE token=?;", (token,), return_num_affected=True
    )
    return num_deleted


def delete_expired_tokens():
    num_deleted = run_sql(
        """
        DELETE FROM token
        WHERE datetime(last_accessed_at, lifespan) < datetime('now');
        """,
        return_num_affected=True,
    )
    return num_deleted



import base64
import json
import re
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import timedelta
from pathlib import Path
from typing import List, Tuple

import requests
from flask import Flask, jsonify, make_response, render_template, request, url_for

from .conf import config
from .decorators import handle_source_site_errors, process_token
from .persistence import (
    create_token,
    delete_token,
    follow,
    get_followed_titles,
    get_prev_next_chapters,
    get_username,
    import_follows,
    load_chapter,
    load_title,
    read,
    register_user,
    save_chapter,
    save_title,
    unfollow,
    unread,
    verify_username_password,
)
from .source_sites import (
    get_chapter,
    get_title,
    search_title_all_sites,
    title_cover,
    title_source_url,
    title_thumbnail,
)
from .storages import storage

config.load()

app = Flask(__name__)
app.config.update(
    SECRET_KEY=config.FLASK_SECRET_KEY,
    PERMANENT_SESSION_LIFETIME=timedelta(days=365),
    MAX_CONTENT_LENGTH=10 * 1024 * 1024,  # max 10MiB payload
)


def _chapter_name(chapter: dict):
    result = ""
    if chapter.get("num_major") is not None:
        result += "Ch. " if chapter.get("volume") else "Chapter "
        result += str(chapter["num_major"])
    if chapter.get("num_minor"):
        result += '.{chapter["num_minor"]}'
    if chapter.get("volume"):
        result += f"Vol. {chapter['volume']}"
    if chapter.get("name"):
        result += f" - {chapter['name']}"
    return result


def _encode_proxy_url(url):
    return base64.urlsafe_b64encode(url.encode()).decode()


def _decode_proxy_url(b64_url):
    return base64.urlsafe_b64decode(b64_url).decode()


def _is_manga_img_url(
    url,
    pattern=re.compile(
        r"^https://("
        r"([\w-]+\.)?(mangadex\.org/(data|images)|mangabeast\d{0,4}.com/manga)/"
        r"|"
        r"([\w-]+\.)+mangadex\.network:\d{2,6}/[\w-]+/data/"
        r")"
    ),
):
    return pattern.match(url)


def proxied(url) -> str:
    path = url_for("proxy_view", b64_url=_encode_proxy_url(url))
    return config.PROXY_PREFIX + path


@app.route("/proxy/<b64_url>")
def proxy_view(b64_url):
    """
    Cached proxy for images (manga cover/page). Motivations:
        - get around source site's hotlinking protection
        - keep working even when source site is down
        - be a polite netizen in general
    """
    url = _decode_proxy_url(b64_url)
    if not _is_manga_img_url(url):
        print("Invalid img url:", url)
        return "Nope", 400

    cached_file_path = Path(config.PROXY_CACHE_DIR) / b64_url
    cached_headers_path = cached_file_path.with_suffix(".headers.json")

    if not (storage.exists(cached_file_path) and storage.exists(cached_headers_path)):
        md_resp = requests.get(url)
        status_code = md_resp.status_code
        body = md_resp.content
        # Normal responsible adults would always include the Content-Type header,
        # but it's MangaDex@Home we're talking about so ofc they wouldn't.
        # Therefore, watch out for that empty case:
        content_type = md_resp.headers.get("content-type")
        headers = {"Content-Type": content_type} if content_type else {}
        if status_code == 200:
            storage.save(cached_headers_path, json.dumps(headers).encode())
            storage.save(cached_file_path, md_resp.content)
    else:
        status_code = 200
        body = storage.read(cached_file_path)
        headers = json.loads(storage.read(cached_headers_path))

    headers["Cache-Control"] = "max-age=31536000"

    return body, status_code, headers


def read_tachiyomi_follows(text: str) -> List[Tuple[str, str]]:
    try:
        data = json.loads(text)
        mangadex_id = None
        mangasee_id = None
        for extension in data["extensions"]:
            id, name = extension.split(":")
            if name == "MangaDex":
                mangadex_id = int(id)
            elif name == "Mangasee":
                mangasee_id = int(id)
        assert mangadex_id and mangasee_id

        results = []
        for manga in data["mangas"]:
            path = manga["manga"][0]
            site_id = manga["manga"][2]
            if site_id == mangadex_id:
                site = "mangadex"
                title_id = path[len("/manga/") : -1]
            elif site_id == mangasee_id:
                site = "mangasee"
                title_id = path[len("/manga/") :]
            else:
                continue
            results.append((site, title_id))

        return results

    except Exception:  # yikes
        import traceback

        traceback.print_exc()
        return None


def ensure_titles(site_title_pairs: List[Tuple[str, str]]):
    """
    Fetch and save titles that are not already in db.
    Returns a list of title dicts, both old and new.
    """
    title_dicts = []
    new_titles = []

    for site, title_id in site_title_pairs:
        existing_title = load_title(site, title_id)
        if existing_title is None:  # again, n+1 queries are fine in sqlite
            new_titles.append((site, title_id))
        else:
            title_dicts.append(existing_title)

    print(f"Fetching {len(new_titles)} new titles out of {len(site_title_pairs)}.")
    with ThreadPoolExecutor(max_workers=3) as executor:
        futures = [
            executor.submit(get_title, site, title_id) for site, title_id in new_titles
        ]
        for future in as_completed(futures):
            title = future.result()
            save_title(title)
            print(f"Saved {title['site']}: {title['name']}")
            title_dicts.append(title)

    return title_dicts


@app.route("/")
@app.route("/h")
@app.route("/a")
@app.route("/f")
@app.route("/s")
@app.route("/s/<query>")
@app.route("/i")
def home_view(query=None):
    return render_template("spa.html")


def _title(site, title_id, user_id=None):
    title = load_title(site, title_id, user_id=user_id)
    if not title:
        print("Getting title", title_id)
        title = get_title(site, title_id)
        print("Saving title", title_id, "to db")
        save_title(title)
    else:
        print("Loading title", title_id, "from db")
    title["cover"] = title_cover(site, title_id, title["cover_ext"])
    if site == "mangadex":
        title["cover"] = proxied(title["cover"])
    title["source_url"] = title_source_url(site, title_id)
    return title


@app.route("/m/<site>/<title_id>")
@handle_source_site_errors("html")
def spa_title_view(site, title_id):
    title = _title(site, title_id)
    return render_template(
        "spa.html",
        open_graph={
            "title": title["name"],
            "image": title["cover"],
            "description": "\n".join(title["descriptions"]),
        },
    )


@app.route("/m/<site>/<title_id>/<chapter_id>")
@handle_source_site_errors("html")
def spa_chapter_view(site, title_id, chapter_id):
    chapter = load_chapter(site, title_id, chapter_id)
    if not chapter:
        print("Getting chapter", chapter_id)
        chapter = get_chapter(site, title_id, chapter_id)
        save_chapter(chapter)
    else:
        print("Loading chapter", chapter_id, "from db")

    # YIIIIKES
    title = load_title(site, title_id)
    title["cover"] = title_cover(site, title_id, title["cover_ext"])
    if site == "mangadex":
        title["cover"] = proxied(title["cover"])

    chapter["site"] = site
    return render_template(
        "spa.html",
        open_graph={
            "title": f'{_chapter_name(chapter)} - {title["name"]}',
            "image": title["cover"],
            "description": "\n".join(title["descriptions"]),
        },
    )


@app.route("/api/title/<site>/<title_id>", methods=["GET"])
@process_token(required=False)
@handle_source_site_errors("json")
def api_title(site, title_id):
    title = _title(site, title_id, user_id=request.user_id)
    return title


@app.route("/api/chapter/<site>/<title_id>/<chapter_id>", methods=["GET"])
@process_token(required=False)
@handle_source_site_errors("json")
def api_chapter(site, title_id, chapter_id):
    chapter = load_chapter(site, title_id, chapter_id)
    if not chapter:
        print("Getting chapter", chapter_id)
        chapter = get_chapter(site, title_id, chapter_id)
        save_chapter(chapter)
    else:
        print("Loading chapter", chapter_id, "from db")

    if site in ("mangadex", "mangasee"):
        chapter["pages"] = [proxied(p) for p in chapter["pages"]]
        chapter["pages_alt"] = [proxied(p) for p in chapter["pages_alt"]]

    # YIIIIKES
    title = load_title(site, title_id)
    title["cover"] = title_cover(site, title_id, title["cover_ext"])
    if site == "mangadex":
        title["cover"] = proxied(title["cover"])
    prev_chapter, next_chapter = get_prev_next_chapters(title, chapter)
    chapter["prev_chapter"] = prev_chapter
    chapter["next_chapter"] = next_chapter
    chapter["site"] = site
    return chapter


@app.route("/api/register", methods=["POST"])
def api_register():
    username = request.json["username"].strip()
    password = request.json["password"]
    message = None
    if not (username and password):
        message = "Empty field(s) spotted. Protip: spaces don't count."
        status_code = 400
    elif (
        len(username) < 2
        or len(username) > 15
        or len(password) < 5
        or len(password) > 50
    ):
        message = "Invalid username/password length. Username length should be 2~15, password 5~50."
        status_code = 400
    else:  # success!
        err = register_user(username, password)
        if err:
            message = err
            status_code = 400
        else:
            message = "Registration successful! You can login now."
            status_code = 200
    return (jsonify({"message": message}), status_code)


@app.route("/api/login", methods=["POST"])
def api_login():
    username = request.json["username"].strip()
    password = request.json["password"]
    remember = request.json["remember"]

    if not (username and password):
        return (
            jsonify({"message": "Empty field(s) spotted. Protip: spaces don't count."}),
            400,
        )

    user_id = verify_username_password(username, password)
    if not user_id:
        return jsonify({"message": "Wrong username/password combination."}), 401

    token = create_token(user_id, remember)
    resp = make_response(jsonify({"user_id": user_id, "token": token}))
    resp.set_cookie(
        "token", token, max_age=31536000 if remember else None, samesite="Strict"
    )
    return resp


@app.route("/api/verify-token", methods=["GET"])
@process_token(required=True)
def api_verify_token():
    return {"user_id": request.user_id, "username": get_username(request.user_id)}, 200


@app.route("/api/logout", methods=["POST"])
@process_token(required=True)
def api_logout():
    num_deleted = delete_token(request.token)
    if num_deleted != 1:
        return jsonify({"message": "Invalid token."}), 401
    resp = make_response("{}", 200)
    resp.set_cookie("token", "", expires=0)
    return resp


@app.route("/api/follows", methods=["GET"])
@process_token(required=True)
def api_follows():
    titles = get_followed_titles(request.user_id)
    for title in titles:
        thumbnail = title_thumbnail(title["site"], title["id"])
        if title["site"] == "mangadex":
            thumbnail = proxied(thumbnail)
        title["thumbnail"] = thumbnail
    return jsonify({"titles": titles})


@app.route("/api/search/<query>", methods=["GET"])
def api_search(query):
    results = search_title_all_sites(query)

    if "mangadex" in results:
        for title in results["mangadex"]:
            title["thumbnail"] = proxied(title["thumbnail"])
    return results


@app.route("/api/follow", methods=["POST"])
@process_token(required=True)
@handle_source_site_errors("json")
def api_follow():
    should_follow = request.json["follow"]
    site = request.json["site"]
    title_id = request.json["title_id"]

    if should_follow:
        follow(request.user_id, site, title_id)
    else:
        unfollow(request.user_id, site, title_id)

    return jsonify({"follow": should_follow})


@app.route("/api/read", methods=["POST"])
@process_token(required=True)
def api_read():
    reads = request.json.get("read") or []
    unreads = request.json.get("unread") or []
    assert reads or unreads

    if reads:
        for r in reads:
            read(
                request.user_id,
                r["site"],
                r["title_id"],
                r["chapter_id"],
            )
    if unreads:
        for u in unreads:
            unread(
                request.user_id,
                u["site"],
                u["title_id"],
                u["chapter_id"],
            )
    # TODO: rewrite read/unread to do bulk updates instead of n+1 queries like these.
    # ... Or maybe not. SQLite doesn't mind.

    # Also TODO: maybe a separate "read all from title" API would be cleaner & easier on
    # FE side.
    return {}


@app.route("/api/import", methods=["POST"])
@process_token(required=True)
@handle_source_site_errors("json")
def api_import():
    # check if the post request has the file part
    if "tachiyomi" not in request.files:
        return jsonify({"message": "No file provided"}), 400
    file = request.files["tachiyomi"]

    # if user does not select file, browser also
    # submits an empty part without filename
    if file.filename == "":
        return jsonify({"message": "No selected file"}), 400

    if file:
        text = file.read()
        site_title_pairs = read_tachiyomi_follows(text)
        if site_title_pairs is None:
            return jsonify({"message": "Malformed file."}), 400

        # First fetch & save titles if they're not already in db
        titles = ensure_titles(site_title_pairs)

        # Then follow them all
        import_follows(request.user_id, site_title_pairs)

        # Mark all chapters as read too
        for title in titles:
            for chapter in title["chapters"]:
                read(request.user_id, title["site"], title["id"], chapter["id"])

        return jsonify({"message": f"Added {len(site_title_pairs)} follows."})



from pytaku.conf import config


def serve():
    """
    This command passes all CLI args to gunicorn, so for example:
        pytaku -w 7 -b 0.0.0.0:5001
    would mean:
        gunicorn pytaku.main:app -w 7 -b 0.0.0.0:5001

    It assumes your virtualenv's bin path is already in $PATH of course.
    """
    import subprocess
    from sys import argv

    command = ["gunicorn", "pytaku.main:app"] + argv[1:]
    print("Running:", " ".join(command))

    subprocess.run(command)


def dev():
    import os
    import subprocess
    from sys import argv

    command = ["flask", "run"] + argv[1:]
    print("Running:", " ".join(command))

    subprocess.run(
        command,
        env={"FLASK_ENV": "development", "FLASK_APP": "pytaku.main:app", **os.environ},
    )


def migrate():
    import argparse

    from .database.migrator import migrate

    argparser = argparse.ArgumentParser()
    argparser.add_argument(
        "-d",
        "--dev",
        action="store_true",
        help="dev mode: overwrites latest_schema.sql on success",
    )
    args = argparser.parse_args()

    config.load()
    migrate(overwrite_latest_schema=args.dev)


def generate_config():
    print(config.generate_json(DEBUG=True))


def scheduler():
    from .scheduler import main_loop

    main_loop()
